import dataclasses
import json
import pathlib

import yaml

import zntrack

CWD = pathlib.Path(__file__).parent.resolve()


@dataclasses.dataclass
class SimpleThermostat:
    """Simple thermostat class"""


@dataclasses.dataclass
class Thermostat:
    temperature: float
    friction: float = 0.1


@dataclasses.dataclass
class Thermostat2:
    temperature: float
    friction: float = 0.1


class MLThermostat(zntrack.Node):
    temp: float = zntrack.params()


class MD(zntrack.Node):
    thermostat: (
        SimpleThermostat
        | Thermostat
        | Thermostat2
        | list[Thermostat | MLThermostat | Thermostat2]
    ) = zntrack.deps()
    steps: int = zntrack.params()


class MD2(zntrack.Node):
    """Need to test two deps"""

    t1: Thermostat = zntrack.deps()
    t2: Thermostat = zntrack.deps()


def test_deps_outside_graph(proj_path):
    thermostat = Thermostat(temperature=300)
    simple_thermostat = SimpleThermostat()

    with zntrack.Project() as project:
        md = MD(thermostat=thermostat, steps=100)

    with project.group("multiple_deps"):
        ml = MLThermostat(temp=300)

        t1 = Thermostat(temperature=300, friction=0.05)
        t2 = Thermostat2(temperature=400)
        md2 = MD(thermostat=[t1, ml, t2], steps=100)

    with project.group("md2"):
        _ = MD2(t1=t1, t2=t2)

    with project.group("md3"):
        _ = MD(thermostat=simple_thermostat, steps=100)

    project.build()

    # TODO: lists / dicts / tuples
    assert md.name == "MD"
    assert md2.name == "multiple_deps_MD"

    node = md.from_rev()
    assert node.thermostat.temperature == thermostat.temperature

    node2 = md2.from_rev(name=md2.name)
    assert isinstance(node2.thermostat[0], Thermostat)
    assert isinstance(node2.thermostat[1], MLThermostat)
    assert isinstance(node2.thermostat[2], Thermostat2)
    assert node2.thermostat[0].temperature == t1.temperature
    assert node2.thermostat[1].temp == ml.temp
    assert node2.thermostat[2].temperature == t2.temperature

    assert json.loads(
        (CWD / "zntrack_config" / "dataclass_deps.json").read_text()
    ) == json.loads((proj_path / "zntrack.json").read_text())
    assert yaml.safe_load(
        (CWD / "dvc_config" / "dataclass_deps.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "dataclass_deps.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()


if __name__ == "__main__":
    test_deps_outside_graph(None)



import zntrack


class HasNumber(zntrack.Node):
    number: int = zntrack.outs()


class CreateNumbers(HasNumber):
    def run(self):
        self.number = 42


class AddOne(HasNumber):
    inp: HasNumber = zntrack.deps()

    def run(self):
        self.number = self.inp.number + 1


class SubtractOne(HasNumber):
    inp: HasNumber = zntrack.deps()

    def run(self):
        self.number = self.inp.number - 1


class Summation(HasNumber):
    """Stage that is actually tested, containing the multiple dependencies"""

    inp: list[HasNumber] = zntrack.deps()

    def run(self):
        self.number = self.inp[0].number + self.inp[1].number


class SummationTuple(HasNumber):
    """Stage that is actually tested, containing the multiple dependencies

    Additionally testing for tuple conversion here!
    """

    inp: list[HasNumber] = zntrack.deps()

    def run(self):
        self.number = self.inp[0].number + self.inp[1].number


def test_repro(proj_path):
    """Test that a single DVC.deps() can take a list of dependencies"""
    project = zntrack.Project()

    with project:
        create_number = CreateNumbers()

        add_one = AddOne(inp=create_number)
        subtract_one = SubtractOne(inp=create_number)

        Summation(inp=[add_one, subtract_one])
        SummationTuple(inp=(add_one, subtract_one))

    project.run()

    assert Summation.from_rev().number == 84
    assert SummationTuple.from_rev().number == 84



import typing as t

import pytest

import zntrack


class AcceptDeps(zntrack.Node):
    deps: t.Any = zntrack.deps()


def test_generic_deps(proj_path):
    with zntrack.Project() as proj:
        AcceptDeps(deps=1)

    with pytest.raises(ValueError):
        proj.build()


def test_generic_deps_list(proj_path):
    with zntrack.Project() as proj:
        AcceptDeps(deps=[1, 2, 3])

    with pytest.raises(ValueError):
        proj.build()



import pathlib
import typing

import zntrack.examples


class SingleNode(zntrack.Node):
    path1: pathlib.Path = zntrack.outs_path()

    def run(self):
        self.path1.write_text("")


class SingleNodeDefaultNWD(zntrack.Node):
    path1: pathlib.Path = zntrack.outs_path(zntrack.nwd / "test.json")

    def run(self):
        self.path1.write_text("")


class SingleNodeListOut(zntrack.Node):
    paths: typing.List[pathlib.Path] | typing.Tuple[pathlib.Path] = zntrack.outs_path()

    def run(self):
        for path in self.paths:
            path.write_text("Lorem Ipsum")


class AssertTempPath(zntrack.Node):
    path1: pathlib.Path = zntrack.outs_path(zntrack.nwd / "test.json")

    def run(self):
        with self.state.use_tmp_path():
            assert self.state.tmp_path is None
            assert self.name == "AssertTempPath"
            assert self.path1 == pathlib.Path("nodes", self.name, "test.json"), self.path1
            self.path1.write_text("lorem ipsum")


def test_run_temp_path(proj_path):
    project = zntrack.Project()
    with project:
        AssertTempPath()
    project.repro()


def test_load_dvc_outs(proj_path):
    with zntrack.Project() as proj:
        node = SingleNode(path1=pathlib.Path("test.json"), name="1500")

    proj.build()

    # node = node.from_rev(name="1500")
    assert node.path1 == pathlib.Path("test.json")

    assert SingleNode.from_rev(name="1500").path1 == pathlib.Path("test.json")


def test_multiple_outs(proj_path):
    with zntrack.Project() as proj:
        SingleNodeListOut(paths=[pathlib.Path("test_1.txt"), pathlib.Path("test_2.txt")])
    proj.build()
    proj.run()

    assert pathlib.Path("test_1.txt").read_text() == "Lorem Ipsum"
    assert pathlib.Path("test_2.txt").read_text() == "Lorem Ipsum"
    assert SingleNodeListOut.from_rev().paths == [
        pathlib.Path("test_1.txt"),
        pathlib.Path("test_2.txt"),
    ]


def test_multiple_outs_tuple(proj_path):
    with zntrack.Project() as proj:
        SingleNodeListOut(paths=(pathlib.Path("test_1.txt"), pathlib.Path("test_2.txt")))
    proj.build()
    proj.run()

    assert pathlib.Path("test_1.txt").read_text() == "Lorem Ipsum"
    assert pathlib.Path("test_2.txt").read_text() == "Lorem Ipsum"
    assert SingleNodeListOut.from_rev().paths == [
        pathlib.Path("test_1.txt"),
        pathlib.Path("test_2.txt"),
    ]


class SingleNodeInNodeDir(zntrack.Node):
    path: pathlib.Path | list[pathlib.Path | str] = zntrack.outs_path(
        zntrack.nwd / "test.json"
    )

    def run(self):
        self.nwd.mkdir(parents=True, exist_ok=True)
        if isinstance(self.path, list):
            [pathlib.Path(x).touch() for x in self.path]
        else:
            self.path.write_text("")


def test_SingleNodeInNodeDir(proj_path):
    with zntrack.Project() as proj:
        node = SingleNodeInNodeDir()
    proj.build()
    proj.run()

    assert node.path == pathlib.Path("nodes", "SingleNodeInNodeDir", "test.json")
    assert node.path.exists()

    result = SingleNodeInNodeDir.from_rev()
    assert result.path == pathlib.Path("nodes", "SingleNodeInNodeDir", "test.json")
    assert result.path.exists()


def test_SingleNodeInNodeDirMulti(proj_path):
    with zntrack.Project() as proj:
        node = SingleNodeInNodeDir(
            path=[zntrack.nwd / "test.json", "file.txt"], name="TestNode"
        )
    proj.build()
    proj.run()

    assert node.path == [pathlib.Path("nodes", "TestNode", "test.json"), "file.txt"]
    assert all(pathlib.Path(x).exists() for x in node.path)

    result = SingleNodeInNodeDir.from_rev(name="TestNode")
    assert result.path == [pathlib.Path("nodes", "TestNode", "test.json"), "file.txt"]
    assert all(pathlib.Path(x).exists() for x in result.path)


def test_SingleNodeDefaultNWD(proj_path):
    with zntrack.Project() as proj:
        n1 = SingleNodeDefaultNWD(name="SampleNode")
        n2 = SingleNodeDefaultNWD()
    proj.build()
    proj.run()

    assert n1.path1 == pathlib.Path("nodes", "SampleNode", "test.json")
    assert n2.path1 == pathlib.Path("nodes", "SingleNodeDefaultNWD", "test.json")

    assert SingleNodeDefaultNWD.from_rev().path1 == pathlib.Path(
        "nodes", "SingleNodeDefaultNWD", "test.json"
    )
    assert SingleNodeDefaultNWD.from_rev(name="SampleNode").path1 == pathlib.Path(
        "nodes", "SampleNode", "test.json"
    )


def test_use_tmp_path(proj_path):
    with zntrack.Project(automatic_node_names=True) as proj:
        node = zntrack.examples.WriteDVCOuts(params="test")
        node2 = zntrack.examples.WriteDVCOutsPath(params="test2")

        node3 = zntrack.examples.WriteDVCOuts(params="test", outs="result.txt")
        node4 = zntrack.examples.WriteDVCOutsPath(
            params="test2", outs=(zntrack.nwd / "data").as_posix()
        )

    proj.build()
    proj.run()

    #     node = node.from_rev(node.name)
    #     node2 = node2.from_rev(node2.name)
    #     node3 = node3.from_rev(node3.name)
    #     node4 = node4.from_rev(node4.name)

    assert node.get_outs_content() == "test"
    assert node2.get_outs_content() == "test2"
    assert node3.get_outs_content() == "test"
    assert node4.get_outs_content() == "test2"

    assert node.outs == pathlib.Path("nodes", "WriteDVCOuts", "output.txt")
    assert node2.outs == pathlib.Path("nodes", "WriteDVCOutsPath", "data")
    assert node3.outs == "result.txt"
    assert isinstance(node4.outs, str)
    assert node4.outs == pathlib.Path("nodes", "WriteDVCOutsPath_1", "data").as_posix()

    #     # DOES THIS EVEN MAKE SENSE?
    #     # IDEA: do only use tmp_path if rev or remote is passed
    with node.state.use_tmp_path():
        assert node.outs == pathlib.Path("nodes", "WriteDVCOuts", "output.txt")
    with node2.state.use_tmp_path():
        assert node2.outs == pathlib.Path("nodes", "WriteDVCOutsPath", "data")
    with node3.state.use_tmp_path():
        assert node3.outs == "result.txt"
        assert isinstance(node3.outs, str)
    with node4.state.use_tmp_path():
        assert (
            node4.outs == pathlib.Path("nodes", "WriteDVCOutsPath_1", "data").as_posix()
        )

    #     # fake remote by passing the current directory
    node = node.from_rev(node.name)
    node2 = node2.from_rev(node2.name)
    node3 = node3.from_rev(node3.name)
    node4 = node4.from_rev(node4.name)

    assert node.get_outs_content() == "test"
    assert node2.get_outs_content() == "test2"
    assert node3.get_outs_content() == "test"
    assert node4.get_outs_content() == "test2"

    assert node.outs == pathlib.Path("nodes", "WriteDVCOuts", "output.txt")
    assert node2.outs == pathlib.Path("nodes", "WriteDVCOutsPath", "data")
    assert node3.outs == "result.txt"
    assert isinstance(node4.outs, str)
    assert node4.outs == pathlib.Path("nodes", "WriteDVCOutsPath_1", "data").as_posix()

    # load the nodes so we can use the tmp_path
    node = node.from_rev(node.name, remote=".")
    node2 = node2.from_rev(node2.name, remote=".")
    node3 = node3.from_rev(node3.name, remote=".")
    node4 = node4.from_rev(node4.name, remote=".")

    with node.state.use_tmp_path():
        assert node.state.tmp_path != pathlib.Path("nodes", "WriteDVCOuts")
        assert node.outs == node.state.tmp_path / "output.txt"
        assert isinstance(node.outs, pathlib.PurePath)
    with node2.state.use_tmp_path():
        assert node2.outs == node2.state.tmp_path / "data"
        assert isinstance(node2.outs, pathlib.PurePath)
    with node3.state.use_tmp_path():
        # no NWD, thus no tmp_path usage possible (as of yet)
        assert node3.outs == "result.txt"
        # assert node3.outs == (node3.state.tmp_path / "result.txt").as_posix()
        assert isinstance(node3.outs, str)
    with node4.state.use_tmp_path():
        assert node4.outs == (node4.state.tmp_path / "data").as_posix()
        assert isinstance(node4.outs, str)


def test_use_tmp_path_multi(proj_path):
    with zntrack.Project(automatic_node_names=True) as proj:
        node = zntrack.examples.WriteMultipleDVCOuts(params=["Lorem", "Ipsum", "Dolor"])

    proj.build()
    proj.run()

    assert node.get_outs_content() == ("Lorem", "Ipsum", "Dolor")

    assert node.outs1 == pathlib.Path("nodes", "WriteMultipleDVCOuts", "output.txt")
    assert node.outs2 == pathlib.Path("nodes", "WriteMultipleDVCOuts", "output2.txt")
    assert node.outs3 == pathlib.Path("nodes", "WriteMultipleDVCOuts", "data")

    with node.state.use_tmp_path():
        assert node.state.tmp_path != pathlib.Path("nodes", "WriteMultipleDVCOuts")
        assert node.outs1 == pathlib.Path("nodes", "WriteMultipleDVCOuts", "output.txt")
        assert node.outs2 == pathlib.Path("nodes", "WriteMultipleDVCOuts", "output2.txt")
        assert node.outs3 == pathlib.Path("nodes", "WriteMultipleDVCOuts", "data")

        assert pathlib.Path(node.outs1).read_text() == "Lorem"
        assert pathlib.Path(node.outs2).read_text() == "Ipsum"
        assert (pathlib.Path(node.outs3) / "file.txt").read_text() == "Dolor"

    node = node.from_rev(remote=".")  # fake remote by passing the current directory

    with node.state.use_tmp_path():
        assert node.state.tmp_path != pathlib.Path("nodes", "WriteMultipleDVCOuts")
        assert node.outs1 == (node.state.tmp_path / "output.txt")
        assert node.outs2 == (node.state.tmp_path / "output2.txt")
        assert node.outs3 == (node.state.tmp_path / "data")

        assert pathlib.Path(node.outs1).read_text() == "Lorem"
        assert pathlib.Path(node.outs2).read_text() == "Ipsum"
        assert (pathlib.Path(node.outs3) / "file.txt").read_text() == "Dolor"


def test_use_tmp_path_sequence(proj_path):
    with zntrack.Project(automatic_node_names=True) as proj:
        node = zntrack.examples.WriteDVCOutsSequence(
            params=["Lorem", "Ipsum", "Dolor"],
            outs=[zntrack.nwd / x for x in ["output.txt", "output2.txt", "output3.txt"]],
        )

    proj.build()
    proj.run()

    assert node.outs == [
        pathlib.Path("nodes", "WriteDVCOutsSequence", "output.txt"),
        pathlib.Path("nodes", "WriteDVCOutsSequence", "output2.txt"),
        pathlib.Path("nodes", "WriteDVCOutsSequence", "output3.txt"),
    ]

    for outs in node.outs:
        assert pathlib.Path(outs).exists()

    with node.state.use_tmp_path():
        for outs in node.outs:
            assert pathlib.Path(outs).exists()
            assert pathlib.Path(outs).read_text() in ("Lorem", "Ipsum", "Dolor")
            assert pathlib.Path(outs).parent == pathlib.Path(
                "nodes", "WriteDVCOutsSequence"
            )

    assert node.get_outs_content() == ["Lorem", "Ipsum", "Dolor"]

    node = node.from_rev(remote=".")  # fake remote by passing the current directory

    with node.state.use_tmp_path():
        for outs in node.outs:
            assert pathlib.Path(outs).exists()
            assert pathlib.Path(outs).read_text() in ("Lorem", "Ipsum", "Dolor")
            assert pathlib.Path(outs).parent == node.state.tmp_path

    assert node.get_outs_content() == ["Lorem", "Ipsum", "Dolor"]


# def test_use_tmp_path_exp(tmp_path_2):
#     with zntrack.Project(automatic_node_names=True) as proj:
#         node = zntrack.examples.WriteDVCOuts(params="test")

#     proj.build()
#     proj.run()

#     with proj.create_experiment() as exp1:
#         node.params = "test1"

#     with proj.create_experiment() as exp2:
#         node.params = "test2"

#     proj.run_exp()

#     exp1.load()
#     node1 = exp1["WriteDVCOuts"]
#     assert node1.get_outs_content() == "test1"

#     with node1.state.use_tmp_path():
#         assert node1.outs == node1.state.tmp_path / "output.txt"
#         assert isinstance(node1.outs, pathlib.PurePath)
#         assert pathlib.Path(node1.outs).read_text() == "test1"

#     exp2.load()
#     node2 = exp2["WriteDVCOuts"]
#     assert node2.get_outs_content() == "test2"

#     with node2.state.use_tmp_path():
#         assert node2.outs == node2.state.tmp_path / "output.txt"
#         assert isinstance(node2.outs, pathlib.PurePath)
#         assert pathlib.Path(node2.outs).read_text() == "test2"

#     assert node.get_outs_content() == "test"
#     assert node.outs == pathlib.Path("nodes", "WriteDVCOuts", "output.txt")

#     with node.state.use_tmp_path():
#         assert node.outs == pathlib.Path("nodes", "WriteDVCOuts", "output.txt")
#         assert pathlib.Path(node.outs).read_text() == "test"



import pathlib
import typing as t

import git
import pytest

import zntrack.examples

# from zntrack.project import Experiment
# from zntrack.utils import config


class ZnNodesNode(zntrack.Node):
    """Used zn.nodes"""

    params: t.Any = zntrack.deps()
    result: t.Any = zntrack.outs()

    def run(self) -> None:
        self.result = self.params


@pytest.mark.xfail(reason="pending implementation")
@pytest.mark.parametrize("assert_before_exp", [True, False])
def test_WriteIO(tmp_path_2, assert_before_exp):
    """Test the WriteIO node."""
    with zntrack.Project() as project:
        node = zntrack.examples.ParamsToOuts(params="Hello World")

    project.run()
    # node.load()
    if assert_before_exp:
        assert node.outs == "Hello World"

    # write a non-tracked file using pathlib
    pathlib.Path("test.txt").write_text("Hello World")

    with project.create_experiment(name="exp1") as exp1:
        node.params = "Hello World"

    # check that the file is still there
    assert pathlib.Path("test.txt").read_text() == "Hello World"

    with project.create_experiment(name="exp2") as exp2:
        node.params = "Lorem Ipsum"

    assert exp1.name == "exp1"
    assert exp2.name == "exp2"

    assert project.experiments.keys() == {"exp1", "exp2"}

    # assert isinstance(project.experiments["exp1"], Experiment)

    project.run_exp()
    assert node.from_rev(rev="exp1").params == "Hello World"
    assert node.from_rev(rev="exp1").outs == "Hello World"

    assert node.from_rev(rev="exp2").params == "Lorem Ipsum"
    assert node.from_rev(rev="exp2").outs == "Lorem Ipsum"

    exp2.apply()
    assert (
        zntrack.from_rev("ParamsToOuts").params
        == zntrack.from_rev("ParamsToOuts", rev=exp2.name).params
    )
    exp1.apply()
    assert (
        zntrack.from_rev("ParamsToOuts").params
        == zntrack.from_rev("ParamsToOuts", rev=exp1.name).params
    )


@pytest.mark.xfail(reason="pending implementation")
@pytest.mark.parametrize("assert_before_exp", [True, False])
def test_WriteIO_no_name(tmp_path_2, assert_before_exp):
    """Test the WriteIO node."""
    with zntrack.Project() as project:
        node = zntrack.examples.ParamsToOuts(params="Hello World")

    project.run()
    # node.load()
    if assert_before_exp:
        assert node.outs == "Hello World"

    with project.create_experiment() as exp1:
        node.params = "Hello World"

    with project.create_experiment() as exp2:
        node.params = "Lorem Ipsum"

    project.run_exp()

    exp1.load()
    assert exp1.nodes["ParamsToOuts"].params == "Hello World"
    assert exp1.nodes["ParamsToOuts"].outs == "Hello World"

    assert exp1["ParamsToOuts"].params == "Hello World"
    assert exp1["ParamsToOuts"].outs == "Hello World"

    exp2.load()
    assert exp2.nodes["ParamsToOuts"].params == "Lorem Ipsum"
    assert exp2.nodes["ParamsToOuts"].outs == "Lorem Ipsum"

    assert exp2["ParamsToOuts"].params == "Lorem Ipsum"
    assert exp2["ParamsToOuts"].outs == "Lorem Ipsum"

    assert zntrack.from_rev("ParamsToOuts", rev=exp1.name).params == "Hello World"
    assert zntrack.from_rev("ParamsToOuts", rev=exp1.name).outs == "Hello World"

    assert zntrack.from_rev("ParamsToOuts", rev=exp2.name).params == "Lorem Ipsum"
    assert zntrack.from_rev("ParamsToOuts", rev=exp2.name).outs == "Lorem Ipsum"


def test_project_remove_graph(proj_path):
    with zntrack.Project() as project:
        node = zntrack.examples.ParamsToOuts(params="Hello World")
    project.run()

    assert node.outs == "Hello World"

    with zntrack.Project(remove_existing_graph=True) as project:
        node2 = zntrack.examples.ParamsToOuts(params="Lorem Ipsum", name="node2")
    project.run()

    assert node2.outs == "Lorem Ipsum"


def test_project_repr_node(tmp_path_2):
    with zntrack.Project():
        node = zntrack.examples.ParamsToOuts(params="Hello World")
        print(node)


@pytest.mark.xfail(reason="pending implementation")
def test_automatic_node_names_False(tmp_path_2):
    with pytest.raises(zntrack.exceptions.DuplicateNodeNameError):
        with zntrack.Project(automatic_node_names=False):
            _ = zntrack.examples.ParamsToOuts(params="Hello World")
            _ = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
    with pytest.raises(zntrack.exceptions.DuplicateNodeNameError):
        with zntrack.Project(automatic_node_names=False):
            _ = zntrack.examples.ParamsToOuts(params="Hello World", name="NodeA")
            _ = zntrack.examples.ParamsToOuts(params="Lorem Ipsum", name="NodeA")


def test_automatic_node_names_default(tmp_path_2):
    with zntrack.Project(automatic_node_names=False):
        _ = zntrack.examples.ParamsToOuts(params="Hello World")
        _ = zntrack.examples.ParamsToOuts(params="Lorem Ipsum", name="WriteIO2")


@pytest.mark.xfail(reason="pending implementation")
def test_automatic_node_names_True(tmp_path_2):
    with zntrack.Project(automatic_node_names=True) as project:
        node = zntrack.examples.ParamsToOuts(params="Hello World")
        node2 = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
        assert node.name == "ParamsToOuts"
        assert node2.name == "ParamsToOuts_1"
    project.run()

    with project:
        node3 = zntrack.examples.ParamsToOuts(params="Dolor Sit")
        assert node3.name == "ParamsToOuts_2"

    project.run()

    assert node.name == "ParamsToOuts"
    assert node2.name == "ParamsToOuts_1"
    assert node3.name == "ParamsToOuts_2"

    project.run()
    # project.load()
    assert "ParamsToOuts" in project.nodes
    assert "ParamsToOuts_1" in project.nodes
    assert "ParamsToOuts_2" in project.nodes

    assert node.outs == "Hello World"
    assert node2.outs == "Lorem Ipsum"
    assert node3.outs == "Dolor Sit"


def test_group_nodes(proj_path):
    with zntrack.Project(automatic_node_names=True) as project:
        with project.group() as group_1:
            node_1 = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
            node_2 = zntrack.examples.ParamsToOuts(params="Dolor Sit")
        with project.group() as group_2:
            node_3 = zntrack.examples.ParamsToOuts(
                params="Amet Consectetur", name="Node01"
            )
            node_4 = zntrack.examples.ParamsToOuts(
                params="Adipiscing Elit", name="Node02"
            )
        with project.group("NamedGrp") as group_3:
            node_5 = zntrack.examples.ParamsToOuts(params="Sed Do")
            node_6 = zntrack.examples.ParamsToOuts(params="Eiusmod Tempor", name="Node03")

        node7 = zntrack.examples.ParamsToOuts(params="Hello World")
        node8 = zntrack.examples.ParamsToOuts(params="How are you?")
        node9 = zntrack.examples.ParamsToOuts(params="I'm fine, thanks!", name="Node04")

    project.repro()

    assert node_1.name == "Group1_ParamsToOuts"
    assert node_2.name == "Group1_ParamsToOuts_1"
    assert node_3.name == "Group2_Node01"
    assert node_4.name == "Group2_Node02"
    assert node_5.name == "NamedGrp_ParamsToOuts"
    assert node_6.name == "NamedGrp_Node03"
    assert node7.name == "ParamsToOuts"
    assert node8.name == "ParamsToOuts_1"
    assert node9.name == "Node04"

    assert group_1.names == ("Group1",)
    assert group_2.names == ("Group2",)
    assert group_3.names == ("NamedGrp",)

    # test nodes in groups
    assert node_1 in group_1
    assert node_2 in group_1
    assert node_3 not in group_1
    assert node_4 not in group_1
    assert len(group_1) == 2
    assert group_1.nwd == pathlib.Path("nodes", "Group1")
    assert node_3 in group_2
    assert node_4 in group_2
    assert node_5 in group_3
    assert node_6 in group_3

    for node in group_1.nodes:
        assert node.state.group == group_1

    for node in group_2.nodes:
        assert node.state.group == group_2

    for node in group_3.nodes:
        assert node.state.group == group_3

    # test node_names in project
    assert node_1.name in group_1
    assert node_2.name in group_1
    assert node_3.name in group_2
    assert node_4.name in group_2
    assert node_5.name in group_3
    assert node_6.name in group_3

    assert node_1.name not in group_2
    assert node_2.name not in group_2
    assert node_3.name not in group_1
    assert node_4.name not in group_1

    # test getitem with node name
    assert group_1[node_1.name] == node_1
    assert group_1[node_2.name] == node_2
    assert group_2[node_3.name] == node_3
    assert group_2[node_4.name] == node_4
    assert group_3[node_5.name] == node_5
    assert group_3[node_6.name] == node_6

    with pytest.raises(KeyError):
        group_1["NodeA"]

    # test iter group
    assert list(group_1) == [node_1, node_2]
    assert list(group_2) == [node_3, node_4]
    assert list(group_3) == [node_5, node_6]

    assert group_1.names == ("Group1",)
    assert group_2.names == ("Group2",)
    assert group_3.names == ("NamedGrp",)

    assert (
        zntrack.examples.ParamsToOuts.from_rev(name="NamedGrp_ParamsToOuts").params
        == "Sed Do"
    )


@pytest.mark.xfail(reason="pending implementation")
def test_build_certain_nodes(tmp_path_2):
    # TODO support passing groups to project.build
    with zntrack.Project(automatic_node_names=True) as project:
        node_1 = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
        node_2 = zntrack.examples.ParamsToOuts(params="Dolor Sit")
    project.build(nodes=[node_1, node_2])
    project.repro()

    assert zntrack.from_rev(node_1).outs == "Lorem Ipsum"
    assert zntrack.from_rev(node_2).outs == "Dolor Sit"

    node_1.params = "ABC"
    node_2.params = "DEF"

    project.build(nodes=[node_1])
    project.repro()

    assert zntrack.from_rev(node_1).outs == "ABC"
    assert zntrack.from_rev(node_2).outs == "Dolor Sit"

    project.run(nodes=[node_2])

    assert zntrack.from_rev(node_1).outs == "ABC"
    assert zntrack.from_rev(node_2).outs == "DEF"


@pytest.mark.xfail(reason="pending implementation")
def test_build_groups(tmp_path_2):
    with zntrack.Project(automatic_node_names=True) as project:
        with project.group() as group_1:
            node_1 = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
            node_2 = zntrack.examples.ParamsToOuts(params="Dolor Sit")
        with project.group() as group_2:
            node_3 = zntrack.examples.ParamsToOuts(params="Amet Consectetur")
            node_4 = zntrack.examples.ParamsToOuts(params="Adipiscing Elit")

    project.run(nodes=[group_1])

    assert zntrack.from_rev(node_1).outs == "Lorem Ipsum"
    assert zntrack.from_rev(node_2).outs == "Dolor Sit"

    with pytest.raises(ValueError):
        zntrack.from_rev(node_3)
    with pytest.raises(ValueError):
        zntrack.from_rev(node_4)

    node_2.params = "DEF"

    project.run(nodes=[group_2, node_2])

    assert zntrack.from_rev(node_1).outs == "Lorem Ipsum"

    assert zntrack.from_rev(node_2).outs == "DEF"
    assert zntrack.from_rev(node_3).outs == "Amet Consectetur"
    assert zntrack.from_rev(node_4).outs == "Adipiscing Elit"

    with pytest.raises(TypeError):
        project.run(nodes=42)

    with pytest.raises(ValueError):
        project.run(nodes=[42])

    # assert that the only directories in "nodes/" are "Group1" and "Group2"
    assert {path.name for path in (tmp_path_2 / "nodes").iterdir()} == {
        "Group1",
        "Group2",
    }


@pytest.mark.xfail(reason="pending implementation")
def test_groups_nwd(tmp_path_2):
    with zntrack.Project(automatic_node_names=True) as project:
        node_1 = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
        with project.group():
            node_2 = zntrack.examples.ParamsToOuts(params="Dolor Sit")
        with project.group("CustomGroup"):
            node_3 = zntrack.examples.ParamsToOuts(params="Adipiscing Elit")

    project.build()

    assert node_1.nwd == pathlib.Path("nodes", node_1.name)
    assert node_2.nwd == pathlib.Path(
        "nodes", "Group1", node_2.name.replace("Group1_", "")
    )
    assert node_3.nwd == pathlib.Path(
        "nodes", "CustomGroup", node_3.name.replace("CustomGroup_", "")
    )
    # now load the Nodes and assert as well

    # assert zntrack.from_rev(node_1).nwd == pathlib.Path("nodes", node_1.name)
    # assert zntrack.from_rev(node_2).nwd == pathlib.Path(
    #     "nodes", "Group1", node_2.name.replace("Group1_", "")
    # )
    # assert zntrack.from_rev(node_3).nwd == pathlib.Path(
    #     "nodes", "CustomGroup", node_3.name.replace("CustomGroup_", "")
    # )

    # with open(config.files.zntrack) as f:
    #     data = json.load(f)
    #     data[node_1.name]["nwd"]["value"] = "test"
    #     data[node_2.name].pop("nwd")

    # with open(config.files.zntrack, "w") as f:
    #     json.dump(data, f)

    # assert zntrack.from_rev(node_1).nwd == pathlib.Path("test")
    # assert zntrack.from_rev(node_2).nwd == pathlib.Path("nodes", node_2.name)
    # assert zntrack.from_rev(node_3).nwd == pathlib.Path(
    #     "nodes", "CustomGroup", node_3.name.replace("CustomGroup_", "")
    # )


@pytest.mark.xfail(reason="pending implementation")
def test_groups_nwd_zn_nodes_a(tmp_path_2):
    node = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
    with zntrack.Project(automatic_node_names=True) as project:
        node_1 = ZnNodesNode(node=node)
        with project.group():
            node_2 = ZnNodesNode(node=node)
        with project.group("CustomGroup"):
            node_3 = ZnNodesNode(node=node)

    assert node_1.name == "ZnNodesNode"
    assert node_1.node.name == "ZnNodesNode+node"

    assert node_2.name == "Group1_ZnNodesNode"
    assert node_2.node.name == "Group1_ZnNodesNode+node"

    assert node_3.name == "CustomGroup_ZnNodesNode"
    assert node_3.node.name == "CustomGroup_ZnNodesNode+node"

    project.run()

    assert zntrack.from_rev(node_1).nwd == pathlib.Path("nodes/ZnNodesNode")
    assert zntrack.from_rev(node_2).nwd == pathlib.Path("nodes", "Group1", "ZnNodesNode")
    assert zntrack.from_rev(node_3).nwd == pathlib.Path(
        "nodes", "CustomGroup", "ZnNodesNode"
    )

    project.load()
    assert node_1.result == "Lorem Ipsum"
    assert node_2.result == "Lorem Ipsum"
    assert node_3.result == "Lorem Ipsum"


@pytest.mark.xfail(reason="pending implementation")
def test_groups_nwd_zn_nodes_b(tmp_path_2):
    node = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
    with zntrack.Project(automatic_node_names=True) as project:
        with project.group():
            node_2 = ZnNodesNode(node=node)
        with project.group("CustomGroup"):
            node_3 = ZnNodesNode(node=node)

    project.run()

    assert zntrack.from_rev(node_2).nwd == pathlib.Path("nodes", "Group1", "ZnNodesNode")
    assert zntrack.from_rev(node_3).nwd == pathlib.Path(
        "nodes", "CustomGroup", "ZnNodesNode"
    )

    project.load()
    assert node_2.result == "Lorem Ipsum"
    assert node_3.result == "Lorem Ipsum"


def test_reopening_groups(proj_path):
    with zntrack.Project(automatic_node_names=True) as project:
        with project.group("AL0"):
            node_1 = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
            node_2 = zntrack.examples.ParamsToOuts(params="Dolor Sit")
            node_3 = zntrack.examples.ParamsToOuts(params="Amet Consectetur")
        with project.group("AL0"):
            node_4 = zntrack.examples.ParamsToOuts(params="Adipiscing Elit")

    project.run()

    assert node_1.nwd == pathlib.Path("nodes", "AL0", "ParamsToOuts")
    assert node_2.nwd == pathlib.Path("nodes", "AL0", "ParamsToOuts_1")
    assert node_3.nwd == pathlib.Path("nodes", "AL0", "ParamsToOuts_2")

    assert node_4.nwd == pathlib.Path("nodes", "AL0", "ParamsToOuts_3")


@pytest.mark.xfail(reason="pending implementation")
def test_nested_groups(proj_path):
    with zntrack.Project(automatic_node_names=True) as project:
        with project.group("AL0"):
            node_1 = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
        with project.group("AL0", "CPU"):
            node_2 = zntrack.examples.ParamsToOuts(params="Dolor Sit")
        with project.group("AL0", "GPU"):
            node_3 = zntrack.examples.ParamsToOuts(params="Amet Consectetur")

    project.run()
    project.load()

    assert node_1.nwd == pathlib.Path("nodes", "AL0", "ParamsToOuts")
    assert node_2.nwd == pathlib.Path("nodes", "AL0", "CPU", "ParamsToOuts")
    assert node_3.nwd == pathlib.Path("nodes", "AL0", "GPU", "ParamsToOuts")

    assert node_1.outs == "Lorem Ipsum"
    assert node_2.outs == "Dolor Sit"
    assert node_3.outs == "Amet Consectetur"


@pytest.mark.xfail(reason="pending implementation")
def test_nested_groups_direct_enter(proj_path):
    project = zntrack.Project(automatic_node_names=True)

    with project.group("AL0"):
        node_1 = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
    with project.group("AL0", "CPU"):
        node_2 = zntrack.examples.ParamsToOuts(params="Dolor Sit")
    with project.group("AL0", "GPU"):
        node_3 = zntrack.examples.ParamsToOuts(params="Amet Consectetur")

    project.run()
    project.load()

    assert node_1.nwd == pathlib.Path("nodes", "AL0", "ParamsToOuts")
    assert node_2.nwd == pathlib.Path("nodes", "AL0", "CPU", "ParamsToOuts")
    assert node_3.nwd == pathlib.Path("nodes", "AL0", "GPU", "ParamsToOuts")

    assert node_1.outs == "Lorem Ipsum"
    assert node_2.outs == "Dolor Sit"
    assert node_3.outs == "Amet Consectetur"


@pytest.mark.xfail(reason="pending implementation")
def test_group_dvc_outs(proj_path):
    project = zntrack.Project(automatic_node_names=True)

    with project.group("GRP1"):
        node = zntrack.examples.WriteDVCOuts(params="Hello World")

    project.run()

    assert node.outs == pathlib.Path("nodes", "GRP1", "WriteDVCOuts", "output.txt")
    assert zntrack.examples.WriteDVCOuts.from_rev(node.name).outs == pathlib.Path(
        "nodes", "GRP1", "WriteDVCOuts", "output.txt"
    )


# @pytest.mark.xfail(reason="pending implementation")
# it fails with False
@pytest.mark.parametrize("git_only_repo", [True])
def test_git_only_repo(proj_path, git_only_repo):
    with zntrack.Project(git_only_repo=git_only_repo) as project:
        zntrack.examples.ParamsToOuts(params="Lorem Ipsum")

    project.repro()

    # commit everything
    repo = git.Repo()
    repo.git.add(".")
    repo.index.commit("initial commit")

    if git_only_repo:
        # check if node-meta.json is in the repo index
        assert ("nodes/ParamsToOuts/node-meta.json", 0) in repo.index.entries.keys()
    else:
        # check if node-meta.json is not in the repo index
        assert ("nodes/ParamsToOuts/node-meta.json", 0) not in repo.index.entries.keys()


@pytest.mark.xfail(reason="pending implementation")
def test_auto_remove(proj_path):
    with zntrack.Project(automatic_node_names=True) as project:
        n1 = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
        n2 = zntrack.examples.ParamsToOuts(params="Dolor Sit")

    project.run()

    n1 = zntrack.examples.ParamsToOuts.from_rev(n1.name)
    n2 = zntrack.examples.ParamsToOuts.from_rev(n2.name)
    assert n1.outs == "Lorem Ipsum"
    assert n2.outs == "Dolor Sit"

    repo = git.Repo()
    repo.git.add(".")
    repo.index.commit("initial commit")

    with zntrack.Project(automatic_node_names=True) as project:
        n1 = zntrack.examples.ParamsToOuts(params="Hello World")

    project.run(auto_remove=True)

    n1 = zntrack.examples.ParamsToOuts.from_rev(n1.name)
    # with pytest.raises(zntrack.exceptions.NodeNotAvailableError):
    #     n2 = zntrack.examples.ParamsToOuts.from_rev(n2.name)


@pytest.mark.xfail(reason="pending implementation")
def test_magic_names(proj_path):
    node = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
    assert node.name == "ParamsToOuts"
    with pytest.raises(ValueError):
        project = zntrack.Project(magic_names=True, automatic_node_names=True)

    project = zntrack.Project(magic_names=True, automatic_node_names=False)
    with project:
        node01 = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
        node02 = zntrack.examples.ParamsToOuts(params="Dolor Sit")
        node03 = zntrack.examples.ParamsToOuts(params="Test01")
    assert node01.name == "node01"
    assert node02.name == "node02"
    assert node03.name == "node03"

    with project.group("Grp01"):
        node01 = zntrack.examples.ParamsToOuts(params="Lorem Ipsum")
        node02 = zntrack.examples.ParamsToOuts(params="Dolor Sit")
        grp_node03 = zntrack.examples.ParamsToOuts(params="Test02")

    assert node01.name == "Grp01_node01"
    assert node02.name == "Grp01_node02"
    assert grp_node03.name == "Grp01_grp_node03"

    project.run()

    assert zntrack.from_rev(node01.name).outs == "Lorem Ipsum"
    assert zntrack.from_rev(node02.name).outs == "Dolor Sit"
    assert zntrack.from_rev(node03.name).outs == "Test01"
    assert zntrack.from_rev(grp_node03.name).outs == "Test02"


def test_group_names(proj_path):
    project = zntrack.Project()
    with project.group() as grp1:
        pass
    with project.group() as grp2:
        pass
    with project.group("NamedGrp") as grp3:
        pass

    assert grp1.names == ("Group1",)
    assert grp2.names == ("Group2",)
    assert grp3.names == ("NamedGrp",)



import pytest
import yaml
from typer.testing import CliRunner

import zntrack
import zntrack.examples
from zntrack import utils
from zntrack.cli import app


@pytest.fixture()
def runner() -> CliRunner:
    return CliRunner()


def test_version(runner):
    result = runner.invoke(app, ["--version"])
    assert "ZnTrack " in result.stdout  # ZnTrack v.x.y


def test_run(proj_path, runner):
    with zntrack.Project() as proj:
        node = zntrack.examples.ParamsToOuts(params=15)

    proj.build()
    result = runner.invoke(app, ["run", "zntrack.examples.ParamsToOuts"])
    assert result.exit_code == 0

    assert node.outs == 15


def test_run_w_name(proj_path, runner):
    with zntrack.Project() as proj:
        node = zntrack.examples.ParamsToOuts(params=15, name="TestNode")

    proj.build()

    result = runner.invoke(
        app, ["run", "zntrack.examples.ParamsToOuts", "--name", "TestNode"]
    )
    assert result.exit_code == 0

    assert node.outs == 15


def test_list_groups(proj_path, runner):
    with zntrack.Project() as proj:
        _ = zntrack.examples.ParamsToOuts(params=15)
        _ = zntrack.examples.ParamsToOuts(params=15)

    with proj.group("example1"):
        _ = zntrack.examples.ParamsToOuts(params=15)
        _ = zntrack.examples.ParamsToOuts(params=15)

    with proj.group("nested", "GRP1"):
        _ = zntrack.examples.ParamsToOuts(params=15)
        _ = zntrack.examples.ParamsToOuts(params=15)
    with proj.group("nested", "GRP2"):
        _ = zntrack.examples.ParamsToOuts(params=15)
        _ = zntrack.examples.ParamsToOuts(params=15)

    proj.build()

    true_groups = {
        "example1": [
            "ParamsToOuts -> example1_ParamsToOuts",
            "ParamsToOuts_1 -> example1_ParamsToOuts_1",
        ],
        "nodes": [
            "ParamsToOuts",
            "ParamsToOuts_1",
        ],
        "nested": [
            {
                "GRP1": [
                    "ParamsToOuts -> nested_GRP1_ParamsToOuts",
                    "ParamsToOuts_1 -> nested_GRP1_ParamsToOuts_1",
                ],
                "GRP2": [
                    "ParamsToOuts -> nested_GRP2_ParamsToOuts",
                    "ParamsToOuts_1 -> nested_GRP2_ParamsToOuts_1",
                ],
            }
        ],
    }

    groups, _ = utils.cli.get_groups(remote=proj_path, rev=None)
    assert groups == true_groups

    result = runner.invoke(app, ["list", proj_path.as_posix()])
    # test stdout == yaml.dump of true_groups
    groups = yaml.safe_load(result.stdout)
    assert groups == true_groups

    assert result.exit_code == 0


def test_list_multi_nested_groups(proj_path, runner):
    proj = zntrack.Project()

    with proj:
        zntrack.examples.ParamsToOuts(params=15)
        zntrack.examples.ParamsToOuts(params=15)

    with proj.group("dynamics"):
        zntrack.examples.ParamsToOuts(params=15)
        zntrack.examples.ParamsToOuts(params=15)

    with proj.group("dynamics", "400K"):
        zntrack.examples.ParamsToOuts(params=15)
        zntrack.examples.ParamsToOuts(params=15)

    with proj.group("dynamics", "400K", "B"):
        zntrack.examples.ParamsToOuts(params=15)
        zntrack.examples.ParamsToOuts(params=15)

    proj.build()

    true_groups = {
        "dynamics": [
            {
                "400K": [
                    {
                        "B": [
                            "ParamsToOuts -> dynamics_400K_B_ParamsToOuts",
                            "ParamsToOuts_1 -> dynamics_400K_B_ParamsToOuts_1",
                        ]
                    },
                    "ParamsToOuts -> dynamics_400K_ParamsToOuts",
                    "ParamsToOuts_1 -> dynamics_400K_ParamsToOuts_1",
                ]
            },
            "ParamsToOuts -> dynamics_ParamsToOuts",
            "ParamsToOuts_1 -> dynamics_ParamsToOuts_1",
        ],
        "nodes": ["ParamsToOuts", "ParamsToOuts_1"],
    }

    groups, _ = utils.cli.get_groups(remote=proj_path, rev=None)
    assert groups == true_groups

    result = runner.invoke(app, ["list", proj_path.as_posix()])
    assert result.exit_code == 0
    groups = yaml.safe_load(result.stdout)
    assert groups == true_groups
    assert result.exit_code == 0
